module Media.Events exposing (onAbort, onCanPlay, onCanPlayThrough, onDurationChange, onEmptied, onEnded, onError, onLoadStart, onLoadSuspend, onLoadedData, onLoadedMetadata, onPaused, onPlaying, onProgress, onSeeked, onSeeking, onStalled, onTimeUpdate, onWaiting)

{-| ###Events

@docs onPlaying, onPaused, onTimeUpdate, onDurationChange, onEnded, onAbort, onSeeked, onSeeking, onLoadStart, onLoadedMetadata, onLoadedData, onLoadSuspend, onEmptied, onWaiting, onStalled, onProgress, onCanPlay, onCanPlayThrough, onError

-}

import Array exposing (fromList, get)
import Html exposing (Attribute)
import Html.Events exposing (on)
import Json.Decode exposing (Decoder, field, map)
import Media.State exposing (State, state)


{-| Triggered when a player finishes seeking
-}
onSeeked : (State -> msg) -> Attribute msg
onSeeked tagger =
    on "seeked" <| target tagger state


{-| Triggered when a player begins seeking
-}
onSeeking : (State -> msg) -> Attribute msg
onSeeking tagger =
    on "seeking" <| target tagger state


{-| Triggered when playback begins
-}
onPlaying : (State -> msg) -> Html.Attribute msg
onPlaying tagger =
    on "playing" <| target tagger state


{-| Triggered when the media is paused
-}
onPaused : (State -> msg) -> Html.Attribute msg
onPaused tagger =
    on "pause" <| target tagger state


{-| Triggered when the currentTime of the player has changed.
** NOTE: This doesn't occur on a per frame basis. For performance reasons,
some browsers only trigger this event every so often, as infrequently as ever 250ms
in some cases. **
-}
onTimeUpdate : (State -> msg) -> Attribute msg
onTimeUpdate tagger =
    on "timeupdate" <| target tagger state


{-| Triggered when the duration of a media file changes.
-}
onDurationChange : (State -> msg) -> Attribute msg
onDurationChange tagger =
    on "durationchange" <| target tagger state


{-| Triggered when playback completes
-}
onEnded : (State -> msg) -> Attribute msg
onEnded tagger =
    on "ended" <| target tagger state


{-| Triggered when the playback of media is aborted (such as restarting it)
-}
onAbort : (State -> msg) -> Attribute msg
onAbort tagger =
    on "abort" <| target tagger state


{-| Triggered when enough data has been loaded to begin playback of the media,
but not enough has yet been loaded to play all the way to the end without pausing to
buffer.
-}
onCanPlay : (State -> msg) -> Attribute msg
onCanPlay tagger =
    on "canplay" <| target tagger state


{-| Triggered when enough data has been loaded to playback the file to the end.
**NOTE: This event does mean the whole media file has been buffered, only that enough
has been buffered that if the current download rate is continued, it will be finished
fast enough that the user will never have to wait.

** NOTE: Also fired on a toggle of play/pause **

-}
onCanPlayThrough : (State -> msg) -> Attribute msg
onCanPlayThrough tagger =
    on "canplaythrough" <| target tagger state


{-| Triggered when the media has become empty, such as if the load task
is invoked.
-}
onEmptied : (State -> msg) -> Attribute msg
onEmptied tagger =
    on "emptied" <| target tagger state


{-| Fired when the loading of the media stalls, that is, when the browser is trying
to fetch media data, but it is not forthcoming.
-}
onStalled : (State -> msg) -> Attribute msg
onStalled tagger =
    on "stalled" <| target tagger state


{-| Triggered when a media Error occurs.
** NOTE: These are errors from the native playr of the browser, not errors
generated by this package. **
-}
onError : (State -> msg) -> Attribute msg
onError tagger =
    on "error" <| target tagger state


{-| Triggered when the media's first frame or sample has finished loading.
-}
onLoadedData : (State -> msg) -> Attribute msg
onLoadedData tagger =
    on "loadeddata" <| target tagger state


{-| Triggered when the media's metadata has been completely loaded.
-}
onLoadedMetadata : (State -> msg) -> Attribute msg
onLoadedMetadata tagger =
    on "loadedmetadata" <| target tagger state


{-| Triggered when the loading of the media has begun.
-}
onLoadStart : (State -> msg) -> Attribute msg
onLoadStart tagger =
    on "loadstart" <| target tagger state


{-| Triggered when the loading of the media is suspended for any reason (including
because it has finished downloading
-}
onLoadSuspend : (State -> msg) -> Attribute msg
onLoadSuspend tagger =
    on "suspend" <| target tagger state


{-| Sent when a task is waiting on the completion of another task (such as Play having
to wait until Seek is complete).
-}
onWaiting : (State -> msg) -> Attribute msg
onWaiting tagger =
    on "waiting" <| target tagger state


{-| Triggered periodically to provide updates on the download progress of
the media.
-}
onProgress : (State -> msg) -> Attribute msg
onProgress tagger =
    on "progress" <| target tagger state


target : (a -> msg) -> Decoder a -> Decoder msg
target tagger decoder =
    map tagger <| field "target" decoder
